from abc import ABC, abstractmethod

from topcast.tts_item import TTSItem


class TTSProviderBase(ABC):
    def __init__(self):
        self.voices = {}

    def get_voice(self, gender):
        return self.voices.get(gender.lower())

    @abstractmethod
    def tts(self, tts_text: TTSItem):
        pass



from topcast.tts_item import TTSItem

from .base import ChatGPTThemeBase


class Summary(ChatGPTThemeBase):
    def __init__(self):
        super().__init__()

        self.system_prompt = "You are an AI language model, and your task is to generate a brief and entertaining summary of a text. Make it interesting and engaging."
        self.messages = [
            {"role": "user", "content": """I have to create a podcast about the Olkiluoto Nuclear Power Plant. I need a short, entertaing and funny summary to the topic. Here are some information about it:

The Olkiluoto Nuclear Power Plant (Finnish: Olkiluodon ydinvoimalaitos, Swedish: Olkiluoto kärnkraftverk) is one of Finland's two nuclear power plants, the other being the two-unit Loviisa Nuclear Power Plant. The plant is owned and operated by Teollisuuden Voima (TVO), a subsidiary of Pohjolan Voima, and is located on Olkiluoto Island, on the shore of the Gulf of Bothnia, in the municipality of Eurajoki in western Finland, about 20 kilometres (12 mi) from the town of Rauma and about 50 kilometres (31 mi) from the city of Pori.

The Olkiluoto plant consists of two boiling water reactors (BWRs), each with a capacity of 890 MW, and one EPR type reactor (unit 3) with a capacity of 1,600 MW.[1] This makes unit 3 currently the most powerful nuclear power plant unit in Europe and the third most powerful globally.[2][3] Construction of unit 3 began in 2005. Commercial operation began on 16 April 2023[4] after being originally scheduled for May 2009.[5][6]"""},
            {"role": "assistant", "content": """
Cancel
Welcome to this week's podcast episode, where we're diving into the world of Finnish nuclear power and giving you a hilarious lowdown on the Olkiluoto Nuclear Power Plant! Nestled on Olkiluoto Island, this bad boy of a power plant churns out electricity like there's no tomorrow. Located just a hop, skip, and a jump away from the towns of Rauma and Pori, it's Finland's hottest (and most radioactive) vacation destination!

Run by the energy whizzes at Teollisuuden Voima (TVO), the Olkiluoto plant boasts two boiling water reactors that'll make your tea in no time, along with one EPR type reactor that's got some serious juice—1,600 MW of capacity, to be exact! In fact, unit 3 is so powerful, it's the European heavyweight champion of nuclear reactors and ranks third worldwide. Talk about a Finnish power trip!

But it wasn't all smooth sailing for this Finnish marvel. Unit 3's construction kicked off in 2005, and it was supposed to be up and running by May 2009. But, like an overambitious DIY project, it took a tad longer than expected, finally getting down to business on April 16, 2023. Better late than never, right?

So, grab your Geiger counters and protective gear, because we're about to embark on an entertaining and electrifying journey into the heart of Finland's Olkiluoto Nuclear Power Plant!"""}
        ]

    def create_content(self, input_content: str):
        return f"""Could you write another summary? Here are some information about it: {input_content}"""

    def extract_content(self, response_content: str):
        return [TTSItem(text=response_content, gender="female")]



import asyncio

from pydub import AudioSegment

from .cutter import Cutter
from .models import AudioItem, AudioLayer, Chapter, ChapterData, Timeline


class Topcaster:
    def __init__(self, timeline: Timeline = []):
        self.timeline = Timeline(timeline=timeline)
        self.topcast = AudioSegment.empty()
        self.already_generated = False

    def add_chapter(self, audio_layers=[], fade_in: int = 1, fade_out: int = 1, padding_start: int = 0, padding_end: int = 0, crossfade: int = 0, volume: float = 1):
        self.already_generated = False

        chapter = Chapter(audio_layers=audio_layers, fade_in=fade_in, fade_out=fade_out,
                          padding_start=padding_start, padding_end=padding_end, crossfade=crossfade, volume=volume)

        self.timeline.timeline.append(chapter)

    def generate(self):
        self.set_length_setter()
        self.add_chapter_data_dict()
        self.open_audio_files()
        asyncio.run(self.generate_text())
        asyncio.run(self.generate_speech())
        self.cut()

        self.already_generated = True
        return self.topcast

    def set_length_setter(self):
        for chapter in self.timeline.timeline:
            have_set_length = False

            for audio_layer in chapter.audio_layers:
                if audio_layer.sets_length:
                    if have_set_length:
                        raise Exception(
                            "Only one audio layer can set the length of a chapter")
                    else:
                        have_set_length = True

            if not have_set_length:
                chapter.audio_layers[0].sets_length = True

    def add_chapter_data_dict(self):
        for chapter in self.timeline.timeline:
            for audio_layer in chapter.audio_layers:
                audio_layer.data = ChapterData(text_list=[], audio_list=[])

    def open_audio_files(self):
        for chapter in self.timeline.timeline:
            for audio_layer in chapter.audio_layers:
                if type(audio_layer.audio) == str:
                    audio_layer.data.raw_audio = AudioSegment.from_file(
                        audio_layer.audio)

    async def generate_text(self):
        tasks = []
        for chapter in self.timeline.timeline:
            for audio_layer in chapter.audio_layers:
                if isinstance(audio_layer.audio, AudioItem):
                    tasks.append(
                        self.generate_text_and_assign(audio_layer)
                    )
        await asyncio.gather(*tasks)

    async def generate_speech(self):
        tasks = []

        for chapter in self.timeline.timeline:
            for audio_layer in chapter.audio_layers:
                if isinstance(audio_layer.audio, AudioItem):
                    tasks.append(self.generate_speech_and_assign(audio_layer))

        await asyncio.gather(*tasks)

    def cut(self):

        topcast = Cutter(self.timeline).cut()
        self.topcast = topcast

    async def generate_text_and_assign(self, audio_layer):
        chatgpt_theme = audio_layer.audio.theme()

        generated_text = await chatgpt_theme.create_chat_completion(audio_layer.audio.content)

        audio_layer.data.text_list = generated_text

    async def generate_speech_and_assign(self, audio_layer):
        provider = audio_layer.audio.tts_provider()
        tasks = []

        for text in audio_layer.data.text_list:
            tasks.append(
                provider.tts(text)
            )

        results = await asyncio.gather(*tasks)
        audio_layer.data.audio_list = results

    def export(self, file_name="topcast.mp3", format="mp3"):
        if not self.already_generated:
            raise Exception(
                "You need to generate the podcast before you can export it")
        self.topcast.export(file_name, format=format)



import asyncio
from io import BytesIO

from elevenlabs.client import AsyncElevenLabs
from pydub import AudioSegment

from ..tts_item import TTSItem
from .base import TTSProviderBase


class ElevenLabs(TTSProviderBase):
    def __init__(self):
        super().__init__()
        self.voices = {
            'male': "George",
            'female': "Laura"
        }
        self.client = AsyncElevenLabs()

    async def tts(self, tts_text: TTSItem):
        voice = self.get_voice(tts_text.gender)

        async def _synthesize_speech():
            audio_generator = await self.client.generate(text=tts_text.text, voice=voice)
            audio_chunks = [chunk async for chunk in audio_generator]
            audio = b''.join(audio_chunks)
            return AudioSegment.from_file(BytesIO(audio))

        return await _synthesize_speech()



import asyncio
from io import BytesIO

from google.cloud import texttospeech
from pydub import AudioSegment

from topcast.tts_item import TTSItem

from .base import TTSProviderBase


class GCP(TTSProviderBase):
    def __init__(self):
        super().__init__()
        self.voices = {
            'male': texttospeech.VoiceSelectionParams(
                language_code="en-US", ssml_gender=texttospeech.SsmlVoiceGender.MALE,
                name="en-US-Neural2-J"
            ),
            'female': texttospeech.VoiceSelectionParams(
                language_code="en-US", ssml_gender=texttospeech.SsmlVoiceGender.FEMALE,
                name="en-US-Neural2-C"
            )
        }

    async def tts(self, tts_text: TTSItem):
        voice = self.get_voice(tts_text.gender)

        def _synthesize_speech():
            client = texttospeech.TextToSpeechClient()

            # Set the audio configuration
            audio_config = texttospeech.AudioConfig(
                audio_encoding=texttospeech.AudioEncoding.LINEAR16
            )

            input_text = texttospeech.SynthesisInput(text=tts_text.text)
            return AudioSegment.from_file(BytesIO(client.synthesize_speech(
                input=input_text, voice=voice, audio_config=audio_config
            ).audio_content))

    # Run the synchronous Text-to-Speech code in a separate thread
        return await asyncio.to_thread(_synthesize_speech)



import asyncio
from io import BytesIO

from gtts import gTTS
from pydub import AudioSegment

from ..tts_item import TTSItem
from .base import TTSProviderBase


class GTTS(TTSProviderBase):
    def __init__(self):
        super().__init__()

    async def tts(self, tts_text: TTSItem):
        return await asyncio.to_thread(self._synthesize_speech, tts_text)

    def _synthesize_speech(self, tts_text: TTSItem):
        mp3_fp = BytesIO()
        tts = gTTS(text=tts_text.text, lang='en')
        tts.write_to_fp(mp3_fp)
        mp3_fp.seek(0)

        return AudioSegment.from_file(mp3_fp, format="mp3")



from typing import Any, List, Optional, Type, Union

from pydantic import BaseModel, ConfigDict, Field, validator
from pydub.audio_segment import AudioSegment

from topcast.chatgpt_themes.base import ChatGPTThemeBase
from topcast.chatgpt_themes.none_theme import NoneTheme
from topcast.tts_item import TTSItem
from topcast.tts_providers.base import TTSProviderBase
from topcast.tts_providers.cgpt import CGPT
from topcast.tts_providers.gt import GTTS


class ChapterData(BaseModel):
    raw_audio: Optional[AudioSegment] = None
    text_list: List[TTSItem] = Field(default_factory=list)
    audio_list: List[AudioSegment] = Field(default_factory=list)

    model_config = ConfigDict(arbitrary_types_allowed=True)


class AudioItem(BaseModel):

    content: str
    theme: Optional[Type[ChatGPTThemeBase]] = Field(default=NoneTheme)
    tts_provider: Optional[Type[TTSProviderBase]] = Field(default=GTTS)

    @validator("theme")
    def check_theme_base_class(cls, value):
        from topcast.chatgpt_themes.base import ChatGPTThemeBase

        if not issubclass(value, ChatGPTThemeBase):
            raise ValueError(
                "The provided theme class does not inherit from ChatGPTThemeBase.")
        return value

    @validator("tts_provider")
    def check_tts_provider_base_class(cls, value):
        from topcast.tts_providers.base import TTSProviderBase

        if not issubclass(value, TTSProviderBase):
            raise ValueError(
                "The provided TTS provider class does not inherit from TTSProviderBase.")
        return value


class AudioLayer(BaseModel):
    audio: Union[str, AudioSegment, AudioItem]
    data: ChapterData = None
    sets_length: bool = False
    fade_in: int = Field(1, ge=1)
    fade_out: int = Field(1, ge=1)
    padding_start: int = Field(0, ge=0)
    padding_end: int = Field(0, ge=0)
    crossfade: int = Field(0, ge=0)
    volume: int = Field(1, ge=0, le=1)

    class Config:
        arbitrary_types_allowed = True


class Chapter(BaseModel):
    audio_layers: List[AudioLayer]
    fade_in: int = Field(1, ge=1)
    fade_out: int = Field(1, ge=1)
    padding_start: int = Field(0, ge=0)
    padding_end: int = Field(0, ge=0)
    crossfade: int = Field(0, ge=0)
    volume: float = Field(1, ge=0, le=1)


class Timeline(BaseModel):
    timeline: List[Chapter]



from abc import ABC, abstractmethod

from openai import OpenAI

client = OpenAI()


class ChatGPTThemeBase(ABC):
    def __init__(self):
        self.system_prompt = ""
        self.messages = []

    def get_messages(self, input_message: str):
        messages = self.messages.copy()
        # insert system prompt at the beginning
        messages.insert(0, {
            "role": "system", "content": self.system_prompt})

        messages.append({
            "role": "user", "content": self.create_content(input_message)})

        return messages

    async def create_chat_completion(self, input_message: str):
        messages = self.get_messages(input_message=input_message)

        completion = client.chat.completions.create(model="gpt-4o",
                                                    messages=messages)
        content = completion.choices[0].message.content

        return self.extract_content(content)

    @abstractmethod
    def create_content(self, input_content: str):
        pass

    @abstractmethod
    def extract_content(self, response_content: str):
        pass



import asyncio
from io import BytesIO

from openai import OpenAI
from pydub import AudioSegment

from ..tts_item import TTSItem
from .base import TTSProviderBase


class CGPT(TTSProviderBase):
    def __init__(self):
        super().__init__()
        self.voices = {
            'male': "onyx",
            'female': "nova"
        }
        self.client = OpenAI()

    async def tts(self, tts_text: TTSItem):
        voice = self.get_voice(tts_text.gender)

        async def _synthesize_speech():
            response = await asyncio.to_thread(
                self.client.audio.speech.create,
                model="tts-1",
                voice=voice,
                input=tts_text.text
            )
            audio_content = response.content
            return AudioSegment.from_file(BytesIO(audio_content), format="mp3")

        return await _synthesize_speech()



from pydub import AudioSegment
from typing import List, Dict, Union
from .models import Timeline, ChapterData, AudioItem

class Cutter:
    def __init__(self, timeline: Timeline):
        self.timeline = timeline

    def cut(self):
        topcast = AudioSegment.empty()
        
        self.set_raw_audio()

        for i, chapter in enumerate(self.timeline.timeline):
            main_audio = None
            overlays = []
            
            for audio_item in chapter.audio_layers:
                if audio_item.sets_length:
                    audio = self.add_effects(audio_item.data.raw_audio, audio_item)
                    main_audio = audio
                
            for audio_item in chapter.audio_layers:
                if not audio_item.sets_length:
                    audio = self.add_effects(audio_item.data.raw_audio[:len(main_audio)], audio_item)
                    overlays.append(audio)
                
            if main_audio is not None:
                chapter_audio = main_audio
                                
                for overlay in overlays:
                    chapter_audio = chapter_audio.overlay(overlay)
                
                chapter_audio = self.add_effects(chapter_audio, chapter)
                
                topcast = topcast.append(chapter_audio, crossfade=chapter.crossfade if i > 0 else 0)

        return topcast
    
    def set_raw_audio(self):
        for chapter in self.timeline.timeline:
            for audio_layer in chapter.audio_layers:
                if not isinstance(audio_layer.data.raw_audio, AudioSegment):
                    raw_audio = AudioSegment.empty()
                    
                    for audio in audio_layer.data.audio_list:
                        raw_audio = raw_audio + audio
                        
                    audio_layer.data.raw_audio = raw_audio

    def add_effects(self, audio: AudioSegment, audio_item: Union[AudioItem, ChapterData]):
        padding_start = AudioSegment.silent(duration=audio_item.padding_start)
        padding_end = AudioSegment.silent(duration=audio_item.padding_end)
        
        audio = padding_start + audio + padding_end

        audio = audio.fade_in(audio_item.fade_in)
        audio = audio.fade_out(audio_item.fade_out)

        # audio = audio.apply_gain(audio_item.volume)
        return audio



from .models import Timeline
from .topcaster import Topcaster


def generate(timeline: Timeline):
    topcaster = Topcaster(timeline)
    topcast = topcaster.generate()
    return topcast


def set_elevenlabs_api_key(api_key: str):
    from elevenlabs import set_api_key
    set_api_key(api_key)


def set_openai_api_key(api_key: str):
    import openai
    openai.api_key = api_key


def set_google_credentials(filename: str):
    import os
    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = filename



from topcast.tts_item import TTSItem

from .base import ChatGPTThemeBase


class NoneTheme(ChatGPTThemeBase):
    def __init__(self):
        pass

    def get_messages(self, input_message: str):
        pass

    async def create_chat_completion(self, input_message: str):
        return [TTSItem(text=input_message, gender="male")]

    def create_content(self, input_content: str):
        pass

    def extract_content(self, response_content: str):
        pass



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="topcast",
    version="0.1.8",
    author="Gorm Labenz",
    author_email="gorm@labenz.io",
    description="A Python package for Topcast",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gormlabenz/topcast",
    packages=find_packages(),
    license="MIT",
    keywords="topcast tts text-to-speech gcp google openai elevenlabs",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.7",
    install_requires=[
        "openai",
        "protobuf",
        "pydantic",
        "pydub",
        "elevenlabs",
        "google-cloud-texttospeech",
        "gTTS",
        "annotated-types",
        "anyio",
        "build",
        "cachetools",
        "certifi",
        "charset-normalizer",
        "click",
        "distro",
        "google-api-core",
        "google-auth",
        "google-cloud-speech",
        "googleapis-common-protos",
        "grpcio",
        "grpcio-status",
        "h11",
        "httpcore",
        "httpx",
        "idna",
        "jiter",
        "packaging",
        "pip-tools",
        "proto-plus",
        "pyasn1",
        "pyasn1_modules",
        "pydantic_core",
        "pyproject_hooks",
        "requests",
        "rsa",
        "sniffio",
        "tqdm",
        "typing_extensions",
        "urllib3",
        "websockets"
    ],
)
